// Imperative and Object-Oriented Programming Methodology
// Uppsala University - Copyright © 2020 Henrik Alfken(Schulze), Lars-Henrik Eriksson

import java.util.Arrays;
import java.util.Objects;

public class Multiset<T extends Comparable<T>> {
  private int nrDistinctElems = 0;
  private int maxCapacity = -1;
  private Integer[] count;
  private T[] elements;

  public Multiset(int maxCapacity) {
    if (maxCapacity < 0) {
      throw new IllegalArgumentException();
    } else {
      this.maxCapacity = maxCapacity;
      count = new Integer[maxCapacity];
      for (int i = 0; i < maxCapacity; i++) {
        count[i] = 0;               // Auto-boxing
      }
      elements = (T[]) new Comparable[maxCapacity];
    }
  }

  // Returns the location in elements where elem exists, alternatively should have existed if the
  // item was included. To be able to distinguish between exists and should have existed, we use
  // negative indexes for the latter, and since 0 is neither positive nor negative, we increase
  // negative indexes by 1. Thus: -5 means that "if the searched element had been in elements
  // it would have been on index 4 "while 4 means" the searched element is in place 4 ".
  // Similarly, -1 means "would have existed first in elements", and 0 "exists first in elements".
  private int indexOf(T elem) {
    if (elem == null) {
      throw new IllegalArgumentException();
    } else {
      int result = -1;
      for (int i = 0; i < nrDistinctElems && result < 0 && elem.compareTo(elements[i]) >= 0; i++) {
        --result;
        if (elem.compareTo(elements[i]) == 0) {
          result = i;
        }
      }
      return result;
    }
  }

  void add(T elem) {
    int indexOf = indexOf(elem);
    if (indexOf < 0 && nrDistinctElems < maxCapacity) {
      int idx = -indexOf - 1;
      // Make room for one more distinct element at index idx:
      for (int j = nrDistinctElems; j > idx; j--) {
        elements[j] = elements[j - 1];
        count[j] = count[j - 1];
      }
      elements[idx] = elem;
      count[idx] = 1;
      ++nrDistinctElems;
    } else if (indexOf > -1) {
      ++count[indexOf];
    } else if (nrDistinctElems == maxCapacity) {
      throw new MultisetFullException();
    }
  }

  public int contains(T elem) {
    int result = 0;
    int indexOf = indexOf(elem);
    if (indexOf > -1) {   // Means that elem DOES exist in elements.
      result = count[indexOf];
    }
    return result;
  }

  public boolean equals(Multiset other) {
    boolean result = false;
    if (this.nrDistinctElems == other.nrDistinctElems) {// Maybe equal, elements must also be equal.
      result = true;
      for (int i = 0; i < nrDistinctElems && result; i++) {
        result = (Objects.equals(this.count[i], other.count[i])); // null-safe equals.
      }
    }
    return result;
  }

  @Override
  public boolean equals(Object other) {
    boolean result;
    if (other instanceof Multiset) {
      result = this.equals((Multiset) other);
    } else {
      result = super.equals(other);
    }
    return result;
  }

  @Override
  public int hashCode() {
    int hash = 3;
    hash = 59 * hash + this.nrDistinctElems;
    hash = 59 * hash + Arrays.deepHashCode(this.count);
    hash = 59 * hash + Arrays.deepHashCode(this.elements);
    return hash;
  }

  public int size() {
    int result = 0;
    for (int i = 0; i < nrDistinctElems; i++) {
      result += count[i];
    }
    return result;
  }
}

// NEVER EVER use default (package private) access. I do it here ONLY to have all code in one file:
class MultisetFullException extends RuntimeException {
  public MultisetFullException() {
    this("");
  }

  public MultisetFullException(String errorMessage) {
    super(errorMessage + "cannot add more elements - the MultiSet is full.");
  }
}